//! # Nostr Protocol Types. Request
//!
//! Used to request events and subscribe to new updates.
use crate::types;
use serde::ser::{self, SerializeSeq, Serializer};

/// Request: used to request events and subscribe to new updates.
///
/// More details at https://github.com/nostr-protocol/nips/blob/master/01.md#communication-between-clients-and-relays
#[derive(Debug, Clone, Default)]
pub struct Request {
    /// The new subscription ID. If not provided, a default one will be created
    /// automatically
    pub subscription_id: types::SubscriptionId,
    /// List of filters to requests events and subscribe to new updates
    /// Events that match any of the filters are to be returned, i.e., multiple filters are to be interpreted as || conditions.
    pub filters: Vec<types::Filter>,
}

impl ser::Serialize for Request {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut seq = serializer.serialize_seq(Some(self.filters.len() + 2))?;
        seq.serialize_element("REQ")?;
        seq.serialize_element(&self.subscription_id)?;
        for element in self.filters.iter() {
            seq.serialize_element(element)?;
        }
        seq.end()
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::Message;
    use chrono::Utc;

    #[test]
    fn encoding_decoding() {
        let r = Request {
            filters: vec![
                types::Filter {
                    authors: vec![
                        "npub1h0h2zfcp3z7acvc9l0qvpgfqznkalj77whsdkmmlwanfssys3ngsx336a8"
                            .try_into()
                            .expect("some addr object"),
                    ],
                    until: Some(Utc::now()),
                    limit: 30,
                    ..Default::default()
                },
                types::Filter {
                    limit: 20,
                    ..Default::default()
                },
            ],
            ..Default::default()
        };

        let serialized = serde_json::to_string(&r).expect("valid json string");
        let obj: Message = serde_json::from_str(&serialized).expect("valid request");
        let obj = obj.as_request().expect("request");
        assert_eq!(r.subscription_id, obj.subscription_id);
        assert_eq!(r.filters.len(), obj.filters.len());
    }
}
