use base64::{engine::general_purpose, Engine as _};
use serde::{
    ser::{self, Serializer},
    Deserialize, Serialize,
};
use std::collections::HashMap;
use thiserror::Error;

use super::Kind;

pub mod profile;

#[derive(Error, Debug)]
pub enum Error {
    #[error("Issue at the serializer: {0}")]
    Json(#[from] serde_json::Error),

    #[error("Base64: {0}")]
    Base64(#[from] base64::DecodeError),
}

#[derive(Eq, PartialEq, Clone, Debug, Serialize, Deserialize)]
pub struct RelaySettings {
    pub read: bool,
    pub write: bool,
}

#[derive(Eq, PartialEq, Clone, Debug)]
pub struct EncryptedData {
    pub encrypted_message: Vec<u8>,
    pub iv: Vec<u8>,
}

#[derive(Eq, PartialEq, Clone, Debug)]
pub enum Content {
    Metadata(profile::Profile),
    ShortTextNote(String),
    RecommendRelayer(String),
    Contacts(HashMap<String, RelaySettings>),
    EncryptedDirectMessage(EncryptedData),
    Unknown(Kind, String),
}

impl ser::Serialize for Content {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        serializer.serialize_str(
            &self
                .try_to_string()
                .map_err(|e| ser::Error::custom(e.to_string()))?,
        )
    }
}

impl Content {
    pub fn kind(&self) -> Kind {
        self.into()
    }

    pub fn deserialize(kind: Kind, content: &str) -> Result<Self, Error> {
        match kind {
            Kind::Metadata => Ok(Self::Metadata(serde_json::from_str(content)?)),
            Kind::ShortTextNote => Ok(Self::ShortTextNote(content.to_owned())),
            Kind::RecommendRelayer => Ok(Self::RecommendRelayer(content.to_owned())),
            Kind::Contacts => Ok(Self::Contacts(serde_json::from_str(content)?)),
            Kind::EncryptedDirectMessages => {
                let parts = content.split("?iv=").collect::<Vec<&str>>();
                if parts.len() == 2 {
                    let encrypted_message = general_purpose::STANDARD.decode(parts[0])?;
                    let iv = general_purpose::STANDARD.decode(parts[1])?;

                    Ok(Self::EncryptedDirectMessage(EncryptedData {
                        encrypted_message,
                        iv,
                    }))
                } else {
                    Ok(Self::Unknown(Kind::Unknown(4), content.to_owned()))
                }
            }
            kind => Ok(Self::Unknown(kind, content.to_owned())),
        }
    }

    pub fn try_to_string(&self) -> Result<String, Error> {
        match self {
            Self::Metadata(p) => Ok(serde_json::to_string(&p)?),
            Self::Contacts(p) => Ok(serde_json::to_string(&p)?),
            Self::EncryptedDirectMessage(message) => {
                let encrypted_message =
                    general_purpose::STANDARD.encode(&message.encrypted_message);
                let iv = general_purpose::STANDARD.encode(&message.iv);
                let to_encode = format!("{}?iv={}", encrypted_message, iv);
                Ok(serde_json::to_string(&to_encode)?)
            }
            Self::ShortTextNote(t) | Self::RecommendRelayer(t) | Self::Unknown(_, t) => {
                Ok(t.to_owned())
            }
        }
    }
}
